#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:47:22
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import Generator, List, Tuple, Union
import requests
import numpy as np
from xagents.model.core import LLM, EMBD, Reranker
from agit.backend.zhipuai_bk import call_llm_api, call_embedding_api
from snippets import batch_process
from xagents.tool.core import ToolCall
from xagents.tool import BaseTool

from loguru import logger

class GLM(LLM):
    def __init__(self, name: str, version: str, api_key=None):
        super().__init__(name, version)
        self.api_key = api_key

    @classmethod
    def list_versions(cls):
        return [
            "glm-4",
            "glm-3-turbo",
            "chatglm3_130b",
            "chatglm_turbo",
            "chatglm_pro",
            "chatglm_66b",
            "chatglm_12b",
            "chatglm2_12b_32k"
        ]

    def _convert_tool_desc(self, tools: List[BaseTool]) -> List[dict]:
        resp = []
        for tool in tools:

            properties = {p.name: dict(type=p.type, description=p.description) for p in tool.parameters}
            required = [p.name for p in tool.parameters if p.required]
            parameters = dict(type="object", properties=properties, required=required)
            tool_desc = dict(type="function", function=dict(name=tool.name, description=tool.description, parameters=parameters))
            resp.append(tool_desc)

        return resp

    def _convert_tool_call(self, tool_calls: List) -> ToolCall:
        tool_call = tool_calls[0]
        tool_call = ToolCall(name=tool_call.function.name, parameters=eval(tool_call.function.arguments), extra_info=dict(tool_call_id=tool_call.id))
        return tool_call

    def observe(self, tool_call: ToolCall,  tools: List[BaseTool] = [], history=[], **kwargs):
        glm_tools = self._convert_tool_desc(tools)
        message = dict(role="tool", content=str(tool_call.resp), tool_call_id=tool_call.extra_info["tool_call_id"])
        resp = call_llm_api(prompt=message, history=history, model=self.version, api_key=self.api_key,
                            tools=glm_tools, logger=logger, return_tool_call=False, **kwargs)
        logger.debug(f"observe result:{resp}")
        return resp

    def generate(self, prompt, history=[], system=None, tools: List[BaseTool] = [], stream=True, **kwargs) -> Tuple[ToolCall, Union[str, Generator]]:
        # logger.info(f"{self.__class__} generating resp with {prompt=}, {history=}")
        glm_tools = self._convert_tool_desc(tools)
        tool_calls, resp = call_llm_api(prompt=prompt, history=history, model=self.version, tools=glm_tools, return_tool_call=True,
                                        do_search=False, system=system, stream=stream, api_key=self.api_key, logger=logger, **kwargs)
        if tool_calls:
            logger.debug(f"tool_calls:{tool_calls}")
            tool_call = self._convert_tool_call(tool_calls)
        else:
            tool_call = None
        return tool_call, resp


class ZhipuEmbedding(EMBD):
    def __init__(self,  api_key=None, batch_size=16, norm=True):
        self.api_key = api_key
        self.batch_size = batch_size
        self.norm = norm

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        # logger.info(f"embedding {len(texts)} with {self.batch_size=}")
        embd_func = batch_process(work_num=self.batch_size, return_list=True)(call_embedding_api)
        embeddings = embd_func(texts, api_key=self.api_key, norm=self.norm)

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        embedding = call_embedding_api(text, api_key=self.api_key, norm=self.norm)
        if isinstance(embedding, np.ndarray):
            embedding = embedding.tolist()

        return embedding

    @classmethod
    def get_dim(cls) -> int:
        return 1024


class ZhipuReranker(Reranker):
    def __init__(self,  url: str, name="reranker", version="ZhipuReranker"):
        self.url = url
        super().__init__(name=name, version=version)

    def cal_similarity(self, text1: str, text2: str):
        logger.debug(f"rerank simi for {text1}, {text2}")
        resp = requests.post(url=self.url, params=dict(text1=text1, text2=text2))
        resp.raise_for_status()
        return resp.json()["data"]["score"]


if __name__ == "__main__":
    # llm_model = GLM(name="glm", version="chatglm_turbo")
    # resp = llm_model.generate("你好", stream=False)
    # print(resp)

    # embd_model = ZhipuEmbedding()
    # text = ["中国", "美国", "日本", "法国", "英国", "意大利", "西班牙", "德国", "俄罗斯"]
    # embds = embd_model.embed_documents(text)
    # print(len(embds))
    # print(embds[0][:4])
    # embd = embd_model.embed_query("你好")
    # print(len(embd))
    # print(embd[:4])

    reranker = ZhipuReranker(url="http://36.103.177.140:8000/get_rel_score", name="zhipu_ranker", version="bge-reranker-base")
    sim = reranker.cal_similarity("私募基金", "公募基金")
    print(sim)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/08 16:57:23
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import copy
import os
from typing import List, Type
from langchain.vectorstores.faiss import FAISS
import faiss
from langchain.vectorstores import VectorStore
from langchain.docstore.in_memory import InMemoryDocstore
from langchain.vectorstores.elasticsearch import ElasticsearchStore
from langchain.vectorstores import VectorStore as VectorStore

from xagents.model.core import EMBD
from xagents.kb.common import DistanceStrategy

from loguru import logger


class XVecStore(VectorStore):
    @classmethod
    def is_local(cls) -> bool:
        raise NotImplementedError()

    @classmethod
    def need_embd(cls):
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config: dict):
        raise NotImplementedError()


class LocalVecStore(VectorStore):
    def __init__(self, local_dir: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.local_dir = local_dir

    @classmethod
    def is_local(cls):
        return True

    @classmethod
    def load_local(cls):
        pass

    def save(self, *args, **kwargs):
        pass


class XFAISS(LocalVecStore, FAISS):
    @classmethod
    def is_local(cls):
        return True

    @classmethod
    def need_embd(cls):
        return False

    @classmethod
    def from_config(cls, config: dict) ->"XFAISS":
        local_dir: str = config["local_dir"]
        embedding: EMBD = config["embedding"]
        distance_strategy = config.get("distance_strategy", DistanceStrategy.MAX_INNER_PRODUCT)

        if os.path.exists(local_dir):
            logger.debug(f"loading vecstore from {local_dir}")
            vecstore = cls.load_local(local_dir=local_dir, embedding=embedding, distance_strategy=distance_strategy)
        else:
            logger.info(f"{local_dir} not exists, create a new local vecstore")

            dim_len = embedding.get_dim()

            # faiss = dependable_faiss_import()
            if distance_strategy == DistanceStrategy.MAX_INNER_PRODUCT:
                index = faiss.IndexFlatIP(dim_len)
            else:
                # Default to L2, currently other metric types not initialized.
                index = faiss.IndexFlatL2(dim_len)
            vecstore = cls(
                local_dir=local_dir,
                index=index, embedding_function=embedding,  index_to_docstore_id=dict(),
                docstore=InMemoryDocstore(), distance_strategy=distance_strategy
            )
        return vecstore

    @classmethod
    def load_local(cls, local_dir:str, embedding:EMBD, distance_strategy:DistanceStrategy,  **kwargs) -> "XFAISS":
        faiss = FAISS.load_local(folder_path=local_dir, allow_dangerous_deserialization=True, embeddings=embedding)
        return XFAISS(local_dir=local_dir, embedding_function=embedding, distance_strategy=distance_strategy,
                       index=faiss.index, docstore=faiss.docstore, index_to_docstore_id=faiss.index_to_docstore_id)

    def save(self, *args, **kwargs):
        return self.save_local(folder_path=self.local_dir, *args, **kwargs)

    def delete(self, ids: List[str], *args, **kwargs) -> int:
        ids = list(set(ids) & set(self.index_to_docstore_id.values()))
        if not ids:
            logger.warning("no overlap id, will not delete from index")
        else:
            logger.debug(f"deleting {len(ids)} vec from index")
            super().delete(ids, *args, **kwargs)
        return len(ids)


class XES(XVecStore, ElasticsearchStore):
    @classmethod
    def is_local(cls):
        return False

    @classmethod
    def need_embd(cls):
        return True

    @classmethod
    def from_config(cls, config: dict):
        vecstore = cls(
            **config
        )
        return vecstore


_vecstores = [XFAISS, XES]
_name2vecstores = {e.__name__: e for e in _vecstores}


def list_vecstores():
    return [e.__name__ for e in _vecstores]


def get_vecstore_cls(name: str) -> Type[XVecStore]:
    return _name2vecstores[name]


def get_vector_store(config: dict, embd_model: EMBD = None) -> XVecStore:
    tmp_config = copy.copy(config)
    vs_cls = tmp_config.pop("vs_cls")
    vs_cls = get_vecstore_cls(vs_cls)
    logger.debug(f"getting vecstore with config:{config}")
    # if vs_cls.need_embd():
    #     if embd_model is None:
    #         raise ValueError("Need embd model to create vector store")
    tmp_config.update(embedding=embd_model)
    return vs_cls.from_documents([], **tmp_config)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 15:39:40
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''



import enum
from pydantic import BaseModel, Field


from typing import List, Tuple

from xagents.loader.common import Chunk
from xagents.config import *
from langchain_core.documents import Document
from snippets import *

def get_kb_dir(kb_name) -> str:
    return os.path.join(KNOWLEDGE_BASE_DIR, kb_name)

def get_chunk_dir(kb_name)->str:
    return os.path.join(get_kb_dir(kb_name), "chunk")
def get_origin_dir(kb_name)->str:
    return os.path.join(get_kb_dir(kb_name), "origin")
def get_id_dir(kb_name)->str:
    return os.path.join(get_kb_dir(kb_name), "id")
def get_index_dir(kb_name)->str:
    return os.path.join(get_kb_dir(kb_name), "index")

def get_config_path(kb_name)->str:
    return os.path.join(get_kb_dir(kb_name), "config.json")

def get_chunk_path(kb_name, file_name) -> str:
    return os.path.join(get_chunk_dir(kb_name), file_name+".jsonl")

def get_origin_path(kb_name, file_name) -> str:
    return os.path.join(get_origin_dir(kb_name), file_name)

def get_id_path(kb_name, file_name) -> str:
    return os.path.join(get_id_dir(kb_name), file_name+".jsonl")


class DistanceStrategy(str, enum.Enum):
    """Enumerator of the Distance strategies for calculating distances
    between vectors."""

    EUCLIDEAN_DISTANCE = "EUCLIDEAN_DISTANCE"
    MAX_INNER_PRODUCT = "MAX_INNER_PRODUCT"


# 知识库中的切片
class KBChunk(Chunk):
    kb_name: str = Field(description="知识库名称")
    file_name: str = Field(description="文件名称")
    idx: int = Field(description="chunk在文档中的顺序,从0开始")

    def to_dict(self):
        return self.model_dump(mode="json", exclude_none=True, exclude={"kb_name", "file_name", "idx"})

    def to_document(self) -> Document:
        if self.search_content:
            page_content, metadata = self.search_content, dict(content=self.content)
        else:
            page_content, metadata = self.content, dict()

        metadata.update(chunk_type=self.content_type.value, idx=self.idx, page_idx=self.page_idx,
                        kb_name=self.kb_name, file_name=self.file_name)
        return Document(page_content=page_content, metadata=metadata)

    @classmethod
    def from_document(cls, document: Document):
        content = document.metadata.pop("content", None)
        item = dict(content=content, search_content=document.page_content) if content else dict(content=document.page_content)
        item.update(document.metadata)
        return cls(**item)

    def __hash__(self) -> int:
        return hash((self.kb_name, self.file_name, self.idx))

    def __eq__(self, __value: object) -> bool:
        return hash(self) == hash(__value)


# 召回的切片
class RecalledChunk(KBChunk):
    query: str = Field(description="召回chunk的query")
    score: float = Field(description="召回chunk的分数")
    forwards: List[Chunk] = Field(description="chunk的下文扩展", default=[])
    backwards: List[Chunk] = Field(description="chunk的上文扩展", default=[])

    @classmethod
    def from_document(cls, document: Document, query: str, score: float) -> "RecalledChunk":
        """从langchain的Document构造过来

        Args:
            document (Document): langchain的Document
            query (str): 相关问题
            score (float): 召回得分

        Returns:
            _type_: RecalledChunk
        """
        chunk = cls.__bases__[0].from_document(document)
        recalled_chunk = cls(**chunk.__dict__, query=query, score=score)
        return recalled_chunk

    def get_content(self):
        if self.search_content:
            return self.search_content + "\n" + self.content

        return self.content

    def to_plain_text(self):
        rs = self.get_content()
        if self.backwards:
            backwards_str = "\n".join([chunk.content for chunk in self.backwards])
            rs = backwards_str + "\n" + rs
        if self.forwards:
            forwards_str = "\n".join([chunk.content for chunk in self.forwards])
            rs = rs+"\n"+forwards_str
        return rs

    def to_detail_text(self, with_context=False, max_len=None) -> str:
        backward_len = sum(len(c.content) for c in self.backwards)
        forwards_len = sum(len(c.content) for c in self.forwards)
        content = self.get_content()
        main_len = len(content)

        detail_text = f"[score={self.score:2.3f}][{main_len}字][扩展后{backward_len+main_len+forwards_len}字][类型{self.content_type.value}][第{self.page_idx}页][index:{self.idx}][相关文档: {self.file_name} ][相关问题:{self.query}]\n\n **{content}**"
        if with_context:
            backwards_str, forwards_str = self.get_contexts(max_len=max_len)
            if backwards_str:
                detail_text = backwards_str + "\n\n"+detail_text
            if forwards_str:
                detail_text = detail_text + "\n\n"+forwards_str
        return detail_text

    def get_contexts(self, max_len=None) -> Tuple[str, str]:
        backwards_str, forwards_str = "", ""
        backward_len = sum(len(c.content) for c in self.backwards)
        forwards_len = sum(len(c.content) for c in self.forwards)

        if backward_len:
            backwards_str = "\n".join([f"{chunk.content}" for idx, chunk in enumerate(self.backwards)])
            if max_len:
                backwards_str = backwards_str[:max_len]+"..."
            backwards_str = f"上文[{backward_len}]字\n\n{backwards_str}"

        if forwards_len:
            forwards_str = "\n".join([f"{chunk.content}" for idx, chunk in enumerate(self.forwards)])
            if max_len:
                forwards_str = forwards_str[:max_len]+"..."
            forwards_str = f"下文[{forwards_len}]字\n\n{forwards_str}"

        return backwards_str, forwards_str


class KnowledgeBaseInfo(BaseModel):
    name: str = Field(description="知识库名称")
    desc: str = Field(description="知识库描述")
    embedding_config: dict = Field(description="embedding模型配置")
    vecstore_config: dict = Field(description="向量存储配置")
    file_num:int = Field(description="知识库文件数量")



# 知识库文件类
class KnowledgeBaseFileInfo(BaseModel):
    kb_name: str = Field(description="知识库名称")
    file_name: str = Field(description="知识库文件名称")
    is_cut:bool = Field(description="是否已经切片")
    is_indexed:bool = Field(description="是否已经索引")





#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/01/04 14:01:01
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import Any, Callable, Dict, List

from pydantic import BaseModel, Field


class Parameter(BaseModel):
    name: str
    type: str
    description: str
    required: bool


class BaseTool(BaseModel):
    name:str = Field(..., description="工具名称")
    description:str = Field(..., description="工具描述")
    parameters: List[Parameter] = Field(..., description="工具参数")
    callable:Callable = Field(..., description="工具执行函数")
    
    
    def execute(self, *args, **kwargs) -> Any:
        return self.callable(*args, **kwargs)

    def to_markdown(self):
        tool_info = f"**[名称]**:{self.name}\n\n**[描述]**:{self.description}\n\n"
        param_infos = []
        for parameter in self.parameters:
            param_infos.append(
                f"- **[参数名]**:{parameter.name}\n\n **[类型]**:{parameter.type}\n\n **[描述]**:{parameter.description}\n\n **[必填]**:{parameter.required}")
        param_infos = "\n\n".join(param_infos)
        return tool_info + param_infos


class ToolCall(BaseModel):
    name: str = Field(description="工具名称")
    parameters: Dict[str, Any] = Field(description="工具调用的参数")
    extra_info: dict = Field(description="额外的信息,比如GLM需要的call_id", default=dict())
    resp: Any = Field(description="工具执行的返回结果,为执行时为None", default=None)

    def to_markdown(self):
        return f"**[调用工具]**: {self.name} **[参数]**: {self.parameters} **[返回结果]**: {self.resp}"



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 18:40:00
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import os

XAGENT_ENV = os.environ.get("XAGENT_ENV", "dev")
XAGENT_HOME = os.path.dirname(os.path.dirname(__file__))

KNOWLEDGE_BASE_DIR = os.environ.get("XAGENT_KNOWLEDGE_BASE_DIR", os.path.join(XAGENT_HOME, "knowledge_base"))
os.makedirs(KNOWLEDGE_BASE_DIR, exist_ok=True)

TEMP_DIR = os.environ.get("XAGENT_TEMP_DIR", os.path.join(XAGENT_HOME, "tmp"))
os.makedirs(TEMP_DIR, exist_ok=True)

LOG_DIR = os.environ.get("XAGENT_LOG_DIR", os.path.join(XAGENT_HOME, "log"))
os.makedirs(LOG_DIR, exist_ok=True)

DB_DIR = os.environ.get("XAGENT_DB_DIR", os.path.join(XAGENT_HOME, "db"))
os.makedirs(DB_DIR, exist_ok=True)


DEFAULT_KB_PROMPT_TEMPLATE = '''请根据[参考信息]回答我的问题，如果问题不在参考信息内，请不要参考
[参考信息]
{context}
问题:
{question}
'''


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 17:04:22
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import os
from typing import List
from xagents.loader.common import AbstractLoader, Chunk, ContentType
from snippets import read2list


class StructedLoader(AbstractLoader):

    def __init__(self, content_key="content",
                 search_content_key="search_content",
                 page_idx_key="page", content_type_key="chunk_type") -> None:
        self.content_key = content_key
        self.page_idx_key = page_idx_key
        self.content_type_key = content_type_key
        self.search_content_key = search_content_key

    @classmethod
    def list_kwargs(cls) -> dict:
        return dict(content_key="content",
                    search_content_key="",
                    page_idx_key="page", content_type_key="chunk_type")

    def load(self, file_path: str) -> List[Chunk]:
        records = read2list(file_path)
        chunks = []
        for item in records[:]:
            rs_item = dict(content=item[self.content_key], page_idx=item.get(self.page_idx_key, 1),
                           content_type=item.get(self.content_type_key, ContentType.TEXT), search_content=item.get(self.search_content_key, None))
            chunks.append(Chunk(**rs_item))
        return chunks


if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    loader = StructedLoader()

    file_path = os.path.join(XAGENT_HOME, "data/raw/贵州茅台2022年报.json")

    loader = StructedLoader(content_key="A", search_content_key="Q")

    file_path = os.path.join(XAGENT_HOME, "data/raw/wind_qa.jsonl")

    chunks = loader.load(file_path=file_path)

    print(len(chunks))
    print(chunks[0])



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/08 14:38:44
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

import copy
import re
from abc import abstractmethod
from typing import Iterable, List

from xagents.config import *
from xagents.kb.common import Chunk
from snippets import batchify
from loguru import logger


class AbstractSplitter:
    def __init__(self, invalid_chunks: List[str] = []):
        self.invalid_chunks = invalid_chunks

    @abstractmethod
    def split(self, text: str) -> List[str]:
        raise NotImplementedError

    def split_chunk(self, chunk: Chunk) -> List[Chunk]:
        rs_chunks = []
        for content in self.split(chunk.content):
            content = content.strip()
            if content in self.invalid_chunks:
                continue
            rs_chunks.append(Chunk(content=content,  content_type=chunk.content_type, page_idx=chunk.page_idx))
        return rs_chunks

# 将文本切分、合并到（min_len~max_len）之间的长度


def merge_cut_texts(texts: Iterable[str], min_len: int, max_len: int) -> Iterable[str]:
    acc = ""
    for text in texts:
        acc += text
        if len(acc) < min_len:
            continue
        if len(acc) > max_len:
            for item in batchify(acc, max_len):
                item = "".join(item)
                if len(item) >= min_len:
                    yield item
                    acc = ""
                else:
                    acc = item
        else:
            yield acc
            acc = ""
    if acc:
        yield acc


class BaseSplitter(AbstractSplitter):

    def __init__(self,
                 separator="\n|。|？|\?|！|！|，",
                 parse_table=False,
                 max_len=100,
                 min_len=5,
                 **kwargs):
        super().__init__(**kwargs)
        self.parse_table = parse_table
        self.separator = separator
        self.max_len = max_len
        self.min_len = min_len

    def _parse_text(self, text: str) -> str:
        text = text.strip()
        text = re.sub("\s+", " ", text)
        text = re.sub(f"\.{3,}", "", text)
        text = text.strip()
        return text

    def split(self, text: str) -> List[str]:
        # logger.info(f"{text=}")
        # logger.info(f"{self.separator=}")
        
        texts = re.split(self.separator, text)
        texts = [self._parse_text(t) for t in texts]
        # logger.debug(f"{texts=}")
        texts = list(merge_cut_texts(texts, self.min_len, self.max_len))
        # logger.debug(f"{text=}")
        return texts

    # def split_chunk(self, chunk: Chunk) -> List[Chunk]:
    #     rs_chunks: List[Chunk] = []
    #     if chunk.content_type == ContentType.TABLE and self.parse_table:
    #         table_chunks = []
    #         tables = markdown2tables(chunk.content)
    #         for table in tables:
    #             descs = table.to_desc()
    #             table_chunks.extend([Chunk(content=e, content_type=ContentType.PARSED_TABLE, page_idx=chunk.page_idx) for e in descs])
    #         rs_chunks.extend(table_chunks)
    #     else:
    #         rs_chunks = super().split_chunk(chunk)
    #         logger.debug(f"{rs_chunks=}")

    #     return rs_chunks


_SPLITTERS = [BaseSplitter]
_NAME2SPLITTER = {s.__name__: s for s in _SPLITTERS}


def get_splitter(config: dict) -> AbstractSplitter:
    tmp_config = copy.copy(config)
    splitter_cls = tmp_config.pop("splitter_cls")
    splitter_cls = _NAME2SPLITTER[splitter_cls]
    return splitter_cls(**tmp_config)


if __name__ == "__main__":
    # splitter = BaseSplitter()
    # texts = ["第一节  释义  ................................ ", " "]
    # for text in texts:
    #     print(splitter.split(text))

    texts = ["a"*4, "b"*5, "b"*4, "c"*51, "d"*4]
    for ele in merge_cut_texts(texts, min_len=5, max_len=25):
        print(f"{ele}, {len(ele)}")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:54:34
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import Generator, List, Optional, Union

from abc import abstractmethod

from pydantic import BaseModel, Field

from xagents.kb.common import RecalledChunk
from xagents.tool.core import ToolCall


class AgentResp(BaseModel):
    content: Union[str, Generator] = Field(description="返回的消息活生成器")
    references: Optional[List[RecalledChunk]] = Field(description="召回的片段")
    tool_call:Optional[ToolCall] = Field(description="工具调用")


class AbstractAgent:

    def __init__(self, name) -> None:
        self.name = name

    @abstractmethod
    def chat(self, message: str, stream=True, do_remember=True) -> AgentResp:
        raise NotImplementedError

    @abstractmethod
    def remember(self, role: str, message: str):
        raise NotImplementedError



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:57:13
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


import os
from typing import Generator, List, Union

from xagents.agent.core import AbstractAgent, AgentResp
from xagents.config import DEFAULT_KB_PROMPT_TEMPLATE, LOG_DIR
from xagents.kb.common import RecalledChunk
from xagents.kb.core import KnowledgeBaseInfo
from xagents.kb.api import get_knowledge_base
from xagents.model.api import get_llm_model
from agent.memory import BaseMemory
from xagents.tool.api import invoke_tool_call
from xagents.tool.core import BaseTool, ToolCall

from loguru import logger

agent_log_path = os.path.join(LOG_DIR, "xagent.log")

# logger.add(agent_log_path, rotation="00:00", retention="7 days", level="INFO", filter=__name__, format=LoguruFormat.SIMPLE)


class XAgent(AbstractAgent):

    def __init__(self, name: str,
                 llm_config: dict,
                 memory_config: dict,
                 tools: List[BaseTool] = [],
                 kb: Union[str, KnowledgeBaseInfo] = None,
                 kb_prompt_template: str = DEFAULT_KB_PROMPT_TEMPLATE) -> None:
        super().__init__(name=name)
        self.llm_model = get_llm_model(llm_config)
        self.memory = BaseMemory(**memory_config)
        if kb:
            if isinstance(kb, str):
                logger.debug(f"loading kb: {kb}")
                self.kb = get_knowledge_base(kb)
            else:
                self.kb = kb
        else:
            self.kb = None
        self.kb_prompt_template = kb_prompt_template
        self.tools = tools

    def search_kb(self, query: str, **kwargs) -> List[RecalledChunk]:
        if not self.kb:
            logger.warning('No knowledge base found! Will return empty recall chunks')
            return []
        chunks = self.kb.search(query=query, **kwargs)
        return chunks

    def use_tool(self, tool_call: ToolCall):
        resp = invoke_tool_call(tool_call)
        return resp

    def chat(self, message: str, stream=True, do_remember=True,
             use_kb=False, top_k=3, score_threshold=None, temperature=0.01,
             do_split_query=False, fake_chat=False, file_names: List[str] = None,
             do_expand=False, expand_len: int = 500, forward_rate: float = 0.5, rerank_config: dict = dict(),
             **kwargs) -> AgentResp:
        logger.info(f"agent get message:{message}")

        if use_kb:
            logger.info("agent searching kb")
            chunks = self.search_kb(query=message, top_k=top_k, score_threshold=score_threshold, do_split_query=do_split_query,
                                    do_expand=do_expand, expand_len=expand_len, forward_rate=forward_rate, file_names=file_names, rerank_config=rerank_config)
            logger.info(f"agent get {len(chunks)} chunks :{chunks}")
            logger.info(f"sample chunks:{chunks[:3]}")
            context = "\n".join(f"{idx+1}." + c.to_plain_text() for idx, c in enumerate(chunks))

            prompt = self.kb_prompt_template.format(question=message, context=context)
        else:
            prompt = message
            chunks = None

        if fake_chat:
            fake_resp = "这是MOCK的回答信息,如果需要真实回答,请设置fake_chat=False"
            model_resp = (e for e in fake_resp) if stream else fake_resp
            tool_call = None
        else:
            tool_call, model_resp = self.llm_model.generate(prompt=prompt, history=self.memory.to_llm_history(), tools=self.tools,
                                                            temperature=temperature, stream=stream, **kwargs)
            # 调用tool
            if tool_call:
                logger.debug(f"calling tool:{tool_call.name}")
                self.use_tool(tool_call)
                history = self.memory.to_llm_history() + [dict(role="user", content=prompt)]
                model_resp = self.llm_model.observe(tool_call, tools=self.tools, history=history,
                                                    stream=stream, temperature=temperature, **kwargs)
                model_resp = "".join(model_resp)
                logger.debug(f"observe tool call resp: {model_resp}")
                model_resp = (e for e in model_resp)

        def _remember_callback(resp_str):
            if do_remember:
                self.remember("user", message)
                self.remember("assistant", resp_str)

        def _add_remember_callback(gen: Generator) -> Generator:
            acc = []
            for ele in gen:
                acc.append(ele)
                yield ele
            resp_str = "".join(acc)
            logger.info(f"agent generate response:{resp_str}")
            _remember_callback("".join(acc))

        if stream:
            model_message = _add_remember_callback(model_resp)
        else:
            model_message = model_resp
            logger.info(f"generate response:{model_message}")
            _remember_callback(model_message)

        resp = AgentResp(content=model_message, references=chunks, tool_call=tool_call)
        return resp

    def remember(self, role: str, message: str):
        logger.debug(f"remembering {role=}, {message=}")
        self.memory.remember(role, message)

    def clear_memory(self):
        # logger.info("clearing memory")
        self.memory.clear()


if __name__ == "__main__":
    llm_config = dict(model_cls="GLM", name="glm", version="chatglm_turbo")
    memory_config = dict(size=10)
    agent = XAgent(name="xagent", llm_config=llm_config, memory_config=memory_config)
    resp = agent.chat("推荐三首歌", stream=True)
    for item in resp:
        print(item, end="")
    resp = agent.chat("其中第二首是谁唱的？", stream=True)
    for item in resp:
        print(item, end="")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/12 15:50:25
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import copy
from typing import List
from loguru import logger
from xagents.model.core import LLM, EMBD, Reranker
from xagents.model.zhipu import ZhipuReranker, ZhipuEmbedding, GLM


_LLM_MODELS = [GLM]
_LLM_MODELS_DICT = {model.__name__: model for model in _LLM_MODELS}


_EMBD_MODELS = [ZhipuEmbedding]
_EMBD_MODELS_DICT = {model.__name__: model for model in _EMBD_MODELS}

_RERANK_MODELS = [ZhipuReranker]
_RERANK_MODELS_DICT = {model.__name__: model for model in _RERANK_MODELS}


def list_llm_models() -> List[str]:
    return [e.__name__ for e in _LLM_MODELS]


def list_embd_models() -> List[str]:
    return [e.__name__ for e in _EMBD_MODELS]


def list_llm_versions(llm_model: str) -> List[str]:
    model_cls = _LLM_MODELS_DICT[llm_model]
    return model_cls.list_versions()


def get_llm_model(config: dict) -> LLM:
    tmp_config = copy.copy(config)
    model_cls = tmp_config.pop("cls")
    model_cls = _LLM_MODELS_DICT[model_cls]
    return model_cls(**tmp_config)


def get_embd_model(config: dict) -> EMBD:
    logger.debug(f"get_embd_model with config:{config}")
    tmp_config = copy.copy(config)
    model_cls = tmp_config.pop("cls")
    if model_cls not in _EMBD_MODELS_DICT:
        msg = f"unknown embedding model: {model_cls}, valid keys:{_EMBD_MODELS_DICT.keys()}"
        raise ValueError(msg)
    model_cls = _EMBD_MODELS_DICT[model_cls]
    return model_cls(**tmp_config)

def get_rerank_model(config:dict)->Reranker:
    if not config:
        return None
    tmp_config = copy.copy(config)
    model_cls = tmp_config.pop("cls")
    model_cls = _RERANK_MODELS_DICT[model_cls]
    return model_cls(**tmp_config)
    


if __name__ == "__main__":
    # config = dict(model_cls="GLM", name="glm", version="chatglm_turbo")
    # model = get_llm_model(config)
    # resp = model.generate(prompt="你好", stream=False)
    # print(resp)
    rerank_config = dict(model_cls="ZhipuReranker", url="http://36.103.177.140:8000/get_rel_score")
    rerank_model = get_rerank_model(rerank_config)
    score = rerank_model.cal_similarity("你好","你滚")
    print(score)
    
    



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/19 16:24:43
@Author  :   ChenHao
@Description  : 文档处理对外接口
@Contact :   jerrychen1990@gmail.com
'''

import os
from loguru import logger
from typing import List, Type
from xagents.loader.pdf import PDFLoader
from xagents.loader.markdown import MarkDownLoader
from xagents.loader.structed import StructedLoader
from xagents.loader.common import Chunk, AbstractLoader

from xagents.loader.splitter import BaseSplitter
from snippets import flat

_EXT2LOADER = {
    "pdf": PDFLoader,
    "markdown": MarkDownLoader,
    "md": MarkDownLoader,
    "json": StructedLoader,
    "jsonl": StructedLoader,
    "csv": StructedLoader,
    "txt": MarkDownLoader,
    "": MarkDownLoader
}


def get_loader_cls(file_path: str) -> Type[AbstractLoader]:
    ext = os.path.splitext(file_path)[-1].lower().replace(".", "")
    loader_cls = _EXT2LOADER[ext]    
    return loader_cls

def load_file(file_path: str, **kwargs) -> List[Chunk]:
    loader_cls = get_loader_cls(file_path)
    loader: AbstractLoader = loader_cls(**kwargs)
    logger.debug(f"loading {file_path} with loader:{loader}")
    contents = loader.load(file_path)
    return contents

def convert2txt(file_path:str, dst_path:str=None, **kwargs)->str:
    """将原始文件转移成

    Args:
        file_path (str): 原始文件路径
        dst_path (str, optional): 目标路径，未传的话，和原始文件同目录. Defaults to None.

    Returns:
        str: 目标路径
    """
    chunks = load_file(file_path, **kwargs)
    if not dst_path:
        dst_path =file_path+".txt"
    with open(dst_path, 'w', encoding='utf-8') as f:
        for chunk in chunks:
            f.write(chunk.content)
    return dst_path







def parse_file(file_path: str,
               separator: str = '\n',
               max_len: int = 200,
               min_len: int = 10) -> List[Chunk]:
    """切分文档，并且按照jsonl格式存储在chunk目录下

    Args:
        kb_name (str): 知识库名称
        file_name (str): 文件名称
        separator (str, optional): 切分符. Defaults to '\n'.
        max_len (int, optional): 最大切片长度. Defaults to 200.
        min_len (int, optional): 最小切片长度. Defaults to 10.

    Returns:
        int: 切片数目
    """
    splitter = BaseSplitter(max_len=max_len, min_len=min_len, separator=separator)

    origin_chunks: List[Chunk] = load_file(file_path=file_path)
    logger.debug(f"load {len(origin_chunks)} origin_chunks")
    
    split_chunks = flat([splitter.split_chunk(origin_chunk) for origin_chunk in origin_chunks])
    logger.info(f"split {len(origin_chunks)} origin_chunks to {len(split_chunks)} chunks")
    return split_chunks


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/15 16:00:04
@Author  :   ChenHao
@Description  :   工具包
@Contact :   jerrychen1990@gmail.com
'''


from functools import wraps
from typing import Callable

from fastapi.params import Body

from loguru import logger


def parse_body(func:Callable):
    @wraps(func)
    def wrapped(**kwargs):
        print(f"{kwargs=}")
        for k,v in kwargs.items():
            if isinstance(v, Body):
                kwargs[k] = v.default
        return func(**kwargs)
    return wrapped


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/19 16:10:35
@Author  :   ChenHao
@Description  :  知识库文件类
@Contact :   jerrychen1990@gmail.com
'''


# 知识库文件类
import os
from loguru import logger
from typing import List
from xagents.kb.common import KBChunk, KnowledgeBaseFileInfo, get_chunk_path, get_id_path, get_origin_path
from xagents.loader.api import parse_file
from xagents.loader.common import Chunk
from snippets import dump, load


class KnowledgeBaseFile:
    def __init__(self, kb_name: str, file_name: str):
        self.kb_name = kb_name
        self.file_name = file_name

        self.chunk_path = get_chunk_path(self.kb_name, self.file_name)
        self.origin_path = get_origin_path(self.kb_name, self.file_name)
        self.id_path = get_id_path(self.kb_name, self.file_name)

    @property
    def is_cut(self) -> bool:
        return os.path.exists(self.chunk_path)

    @property
    def is_indexed(self) -> bool:
        return os.path.exists(self.id_path)

    @property
    def chunk_num(self) -> int:
        if not self.is_cut:
            return 0
        chunks = load(self.chunk_path)
        return len(chunks)

    def get_info(self):
        return KnowledgeBaseFileInfo(kb_name=self.kb_name, file_name=self.file_name, is_cut=self.is_cut, is_indexed=self.is_indexed)

    def cut(self, *args, **kwargs):
        logger.info(f"start cut file: {self.file_name}")
        chunks: List[Chunk] = parse_file(file_path=self.origin_path, *args, **kwargs)
        kb_chunks = [KBChunk(**chunk.model_dump(mode="json"), idx=idx, kb_name=self.kb_name, file_name=self.file_name)
                     for idx, chunk in enumerate(chunks)]
        kb_chunk_json = [chunk.to_dict() for chunk in kb_chunks]
        dump(kb_chunk_json, self.chunk_path)
        return len(kb_chunks)

    def get_chunks(self) -> List[KBChunk]:
        """
        从切片文件加载切片
        """
        if not self.is_cut:
            return []

        chunk_dicts = load(self.chunk_path)
        logger.debug(f"loaded {len(chunk_dicts)} chunks from {self.chunk_path}")
        chunks: List[KBChunk] = [KBChunk(**c, idx=idx, kb_name=self.kb_name, file_name=self.file_name) for idx, c in enumerate(chunk_dicts)]
        return chunks

    def remove(self):
        """
        删除知识库文档
        """
        for path in [self.origin_path, self.chunk_path]:
            if os.path.exists(path):
                os.remove(path)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/03/19 15:45:46
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''
import copy
import os
import re
from typing import List, Type
from loguru import logger
from xagents.kb.kb_file import KnowledgeBaseFile
from xagents.kb.vector_store import LocalVecStore, XVecStore, get_vecstore_cls
from xagents.loader.common import Chunk
from xagents.model.api import get_embd_model, get_rerank_model
from xagents.kb.common import KnowledgeBaseInfo, RecalledChunk, get_chunk_dir, get_chunk_path, get_config_path, get_id_dir, get_index_dir, get_kb_dir, get_origin_dir, DistanceStrategy
from snippets import load, log_cost_time, dump


class KnowledgeBase():

    def __init__(self, name: str, desc, embedding_config: dict, vecstore_config: dict):
        self.name = name
        self.desc = desc
        self.embedding_config = embedding_config
        self.vecstore_config = vecstore_config
        self._build_dirs()
        self._save_config()
        self.index=None

    @classmethod
    def from_config(cls, config: dict | str):
        if isinstance(config, str):
            config = load(config)
        return cls(**config)

    def _get_config(self) -> dict:
        return dict(name=self.name, desc=self.desc, embedding_config=self.embedding_config, vecstore_config=self.vecstore_config)

    def get_info(self) -> KnowledgeBaseInfo:
        """返回知识库的信息

        Returns:
            dict: 知识库信息
        """
        file_num = len(list(self.list_kb_files()))
        return KnowledgeBaseInfo(**self._get_config(), file_num=file_num)

    def _build_dirs(self):
        self.kb_dir = get_kb_dir(self.name)
        self.origin_dir = get_origin_dir(self.name)

        self.chunk_dir = get_chunk_dir(self.name)
        self.index_dir = get_index_dir(self.name)
        self.id_dir = get_id_dir(self.name)
        self.config_path = get_config_path(self.name)

        os.makedirs(self.origin_dir, exist_ok=True)
        os.makedirs(self.chunk_dir, exist_ok=True)
        os.makedirs(self.id_dir, exist_ok=True)

    def _save_config(self):
        dump(self._get_config(), self.config_path)

    def get_index(self) -> XVecStore:
        """
        加载向量库
        """
        if not self.index:
            
            config = copy.copy(self.vecstore_config)
            vecstore_cls: Type[XVecStore] = get_vecstore_cls(config.pop("cls"))
            embedding_model = get_embd_model(self.embedding_config)
            config.update(embedding=embedding_model, local_dir=self.index_dir)
            self.index = vecstore_cls.from_config(config)
        return self.index

    def list_kb_files(self) -> List[KnowledgeBaseFile]:
        kb_files = []
        for file_name in os.listdir(self.origin_dir):
            kb_file = KnowledgeBaseFile(kb_name=self.name, file_name=file_name)
            kb_files.append(kb_file)
        return kb_files

    def remove_file(self, kb_file: KnowledgeBaseFile):
        """
        删除文档
        """
        assert kb_file.kb_name == self.name
        kb_file.remove()
        if kb_file.is_indexed:
            index = self.get_index()
            logger.debug(f"{index=}")
            self.remove_kb_file_from_index(index, kb_file)

    def remove_kb_file_from_index(self, index: XVecStore, kb_file: KnowledgeBaseFile):
        if kb_file.is_indexed:
            doc_ids = load(kb_file.id_path)
            index.delete(doc_ids)
            os.remove(kb_file.id_path)

    def add_kb_file2index(self, index: XVecStore, kb_file: KnowledgeBaseFile, reindex=False, do_save=False):
        if not kb_file.is_cut:
            logger.warning(f"{kb_file.file_name} is not cut, please cut it first")
            return

        if kb_file.is_indexed and not reindex:
            logger.info(f"{kb_file.file_name} is already indexed, skip")
        else:
            self.remove_kb_file_from_index(index, kb_file)
            chunks = kb_file.get_chunks()
            logger.info(f"reindexing {kb_file.file_name} with {len(chunks)}chunks")
            documents = [chunk.to_document() for chunk in chunks]
            # logger.debug(f"sample document:{documents[0]}")
            ids = index.add_documents(documents)
            logger.debug(f"{len(ids)} documents add to index")
            dump(ids, kb_file.id_path)
        # logger.debug(f"{index=}")
        if index.is_local() and do_save:
        # TODO 为何isinstance判断不work？
        # if isinstance(index, LocalVecStore) and do_save:
            logger.info(f"saving {kb_file.file_name} to index dir:{index.local_dir}")
            index.save()

    @log_cost_time(name="rebuild_index")
    def rebuild_index(self, reindex=False):
        """
        重新构建向量知识库
        """
        index = self.get_index()
        kb_files = self.list_kb_files()
        for kb_file in kb_files:
            self.add_kb_file2index(index, kb_file, reindex=reindex, do_save=False)
        if isinstance(index, LocalVecStore):
            index.save()



    @log_cost_time(name="kb_search")
    def search(self, query: str, top_k: int = 3, score_threshold: float = None,
               do_split_query=False, file_names: List[str] = None, rerank_config: dict = {},
               do_expand=False, expand_len: int = 500, forward_rate: float = 0.5) -> List[RecalledChunk]:
        """知识库检索

        Args:
            query (str): 待检索的query
            top_k (int, optional): 返回多少个chunk. Defaults to 3.
            score_threshold (float, optional): 召回的chunk相似度阈值. Defaults to None.
            do_split_query (bool, optional): 是否按照？切分query并分别召回. Defaults to False.
            file_names (List[str], optional): 按照名称过滤需要召回的片段所在的文件. Defaults to None.
            do_expand (bool, optional): 返回的chunk是否做上下文扩展. Defaults to False.
            expand_len (int, optional): 上下文扩展后的chunk字符长度（do_expand=True时生效）. Defaults to 500.
            forward_rate (float, optional): 上下文扩展时向下文扩展的比率（do_expand=True时生效）. Defaults to 0.5.

        Returns:
            List[RecalledChunk]: 相关的切片，按照score降序
        """
        index = self.get_index()
        recalled_chunks = []
        # 切分query
        queries = split_query(query, do_split_query)

        # 将原始score归一化到0-1，越大越接近
        def _get_score(s):
            if self.vecstore_config["distance_strategy"] in [DistanceStrategy.EUCLIDEAN_DISTANCE]:
                return 1.-s
            return s

        # 过滤条件
        _filter = dict()
        if file_names:
            _filter = dict(file_name=file_names)

        # 每个子query做检索
        for query in queries:
            score_threshold = _get_score(score_threshold)
            logger.debug(f"searching {query} with vecstore_cls: {index.__class__.__name__}, {_filter=}, {top_k=}, {score_threshold=}")

            docs_with_score = index.similarity_search_with_score(query, k=top_k, score_threshold=score_threshold, filter=_filter)
            logger.debug(f"{len(docs_with_score)} origin chunks found for {query}")
            logger.debug(f"{query}'s related docs{[d.page_content[:5] for d,s in docs_with_score]}")
            # logger.debug(docs_with_score[0])

            tmp_recalled_chunks = [RecalledChunk.from_document(d, score=_get_score(s), query=query) for d, s in docs_with_score]
            recalled_chunks.extend(tmp_recalled_chunks)

        # 去重，避免召回相同切片
        recalled_chunks = list(set(recalled_chunks))
        logger.info(f"{len(recalled_chunks)} chunks recalled")

        #重排序
        recalled_chunks = rerank(recalled_chunks, rerank_config)[:top_k]
        logger.info(f"get {len(recalled_chunks)} reranked chunks after sort")

        # 上下文扩展
        if do_expand:
            logger.info("expanding recalled chunks")
            for chunk in recalled_chunks:
                expand_chunk(chunk, expand_len, forward_rate)

        return recalled_chunks


def rerank(recalled_chunks: List[RecalledChunk], rerank_config: dict) -> List[RecalledChunk]:
    """重排序
    Args:
        recalled_chunks (List[RecalledChunk]): 待排序的切片

    Returns:
        List[RecalledChunk]: 排序后的切片
    """
    # logger.debug("reranking...")
    rerank_model = get_rerank_model(rerank_config)
    if rerank_model:
        logger.info("reranking chunks with rerank model")
        for chunk in recalled_chunks:
            similarity = rerank_model.cal_similarity(chunk.query, chunk.content)
            chunk.score = similarity

    recalled_chunks.sort(key=lambda x: x.score, reverse=True)
    return recalled_chunks

def split_query(query: str, do_split_query=False) -> List[str]:
    """切割query"""
    if do_split_query:
        rs = [e.strip() for e in re.split("\?|？", query) if e.strip()]
        logger.debug(f"split origin query into {len(rs)} queries")

        return rs
    else:
        # 不需要切分也转成list形式，方便后续统一处理
        return [query]


# 扩展上下文到给定的长度
def expand_chunk(chunk: RecalledChunk, expand_len: int, forward_rate=0.5) -> RecalledChunk:
    logger.debug(f"expanding chunk {chunk}")
    chunk_path = get_chunk_path(chunk.kb_name, chunk.file_name)
    chunks = []
    for item in load(chunk_path):
        tmp = Chunk(**item)
        chunks.append(tmp)
    chunk_idx = chunk.idx

    to_expand = expand_len - len(chunk.content)
    if to_expand <= 0:
        return chunk

    forward_len = int(to_expand * forward_rate)
    backward_len = to_expand - forward_len
    logger.debug(f"expand chunk with :{forward_len=}, {backward_len=}")
    backwards, forwards = [], []

    # 查找前面的chunk
    idx = chunk_idx-1
    while idx >= 0:
        tmp_chunk = chunks[idx]
        backward_len -= len(tmp_chunk.content)
        if backward_len < 0:
            break
        backwards.append(tmp_chunk)
        idx -= 1
    backwards.reverse()

    idx = chunk_idx + 1
    while idx < len(chunks):
        tmp_chunk = chunks[idx]
        forward_len -= len(tmp_chunk.content)
        if forward_len < 0:
            break
        forwards.append(tmp_chunk)
        idx += 1

    logger.debug(f"expand with {len(backwards)} backward chunks and {len(forwards)} forward chunks")
    chunk.backwards = backwards
    chunk.forwards = forwards
    return chunk



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 17:45:15
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from abc import abstractmethod
from typing import Generator, Tuple, Union

from langchain_core.embeddings import Embeddings

from xagents.tool.core import ToolCall

StrOrGen = Union[str, Generator]


class LLM:
    """语言模型的基类
    """
    def __init__(self, name: str, version: str):
        """初始化函数

        Args:
            name (str): LLM的名称
            version (str): 模型版本
        """
        self.name = name
        self.version = version

    @abstractmethod
    def generate(self, prompt:str, history=[], system:str=None, stream=True, **kwargs)->Tuple[ToolCall,StrOrGen]:
        """生成结果

        Args:
            prompt (str): 给LLM的提示词
            history (list, optional): 历史message列表. Defaults to [].
            system (_type_, optional): system信息. Defaults to None.
            stream (bool, optional): 是否返回generator. Defaults to True.
        Returns:
            Tuple[ToolCall,StrOrGen]: ToolCall:工具调用实例，可以为空, StrOrGen:模型回复内容，字符串或generator
        """
        raise NotImplementedError

    @classmethod
    def list_version(cls):
        raise NotImplementedError


class EMBD(Embeddings):
    @classmethod
    def get_dim(cls) -> int:
        raise NotImplementedError

class Reranker:
    def __init__(self, name: str, version: str):
        self.name = name
        self.version = version

    @abstractmethod
    def cal_similarity(self, text1:str, text2:str)->float:
        raise NotImplementedError

    




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/02/20 17:39:31
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from xagents.util import format_prompt

def test_format_prompt():
    prompt = "hello {name}"
    prompt = format_prompt(template=prompt, name="world", age=12)
    print(prompt)
    
    


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/19 13:40:56
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from typing import List, Dict


import requests
from snippets import *
from xagents.model import LLM, EMBD

from loguru import logger
MOCK_ZETA = {'body': {'id': 'chatcmpl-UXqiZMi4rsDdjiwwjHH5CM', 'object': 'chat.completion', 'created': 1701744928, 'model': 'WindAlice-AIGC-70B-dev-20231017', 'choices': [{'index': 0, 'message': {
    'role': 'assistant', 'content': '您好！我是Wind Alice金融智能助手，有什么我可以帮助您的吗？\n\n'}, 'finish_reason': 'stop'}], 'usage': {'prompt_tokens': 16, 'total_tokens': 34, 'completion_tokens': 18}}}


def call_zeta_chat(messages: List[Dict], temperature=0.01, top_p=0.7, max_new_token=1024):
    return MOCK_ZETA
    url = "http://10.100.4.96/zeta-models/chat/zeta66b"
    data = dict(messages=messages, temperature=temperature, top_p=top_p, max_new_token=max_new_token)
    resp = requests.post(url, json=data)
    resp.raise_for_status()
    return resp.json()


def call_zeta_embedding(contents: List[str]):
    import numpy as np
    return np.random.random(size=(len(contents), 1024)).tolist()

    url = "http://10.100.4.96/zeta-models/embedding/zeta"
    data = dict(contents=contents)
    resp = requests.post(url, json=data)
    resp.raise_for_status()
    return resp.json()["embeddings"]


class ZetaEMBD(EMBD):

    def __init__(self,  batch_size=16, norm=True):
        self.batch_size = batch_size
        self.norm = norm

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        logger.info(f"embedding {len(texts)} with {self.batch_size=}")

        embeddings = []
        from tqdm import tqdm
        for idx, batch in tqdm(enumerate(batchify(texts, self.batch_size))):
            try:
                embds = call_zeta_embedding(contents=batch)
                embeddings.extend(embds)
                pct = idx*self.batch_size/len(texts)
                logger.info(f"{idx*self.batch_size}/{len(texts)} [{pct:2.2%}] embeddings done")
            except Exception as e:
                logger.error(f"embedding service error")
                logger.error(batch)

                raise e
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        embedding = call_zeta_embedding(contents=[text])[0]
        return embedding

    @classmethod
    def get_dim(cls) -> int:
        return 1024


class ZetaLLM(LLM):
    @classmethod
    def list_versions(cls):
        return [
            "v1.0.0",
        ]

    def __init__(self, name="ZETA", version="v1.0.0"):
        super().__init__(name=name, version=version)

    def generate(self, prompt, history=[], stream=True,
                 temperature=0.01, **kwargs):
        messages = history + [dict(role="user", content=prompt)]
        response = call_zeta_chat(messages=messages,
                                  temperature=temperature, **kwargs)
        resp_message = response["body"]["choices"][0]["message"]["content"]
        return (e for e in resp_message) if stream else resp_message


if __name__ == "__main__":

    zeta_llm = ZetaLLM()
    resp = zeta_llm.generate("你好", stream=False)
    print(f"llm resp:{resp}")

    zeta_embd = ZetaEMBD()
    contents = ["万得", "金融"]
    embds = zeta_embd.embed_documents(texts=contents)
    print(f"embd num {len(embds)}")
    print(f"embd dim {len(embds[0])}")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 16:40:25
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from abc import abstractmethod
import enum
from typing import List, Optional

from pydantic import BaseModel, Field
from xagents.config import *



# 切片类型


class ContentType(str, enum.Enum):
    TABLE = "TABLE"
    TITLE = "TITLE"
    TEXT = "TEXT"
    PARSED_TABLE = "PARSED_TABLE"

# 切片


class Chunk(BaseModel):
    content: str = Field(description="chunk的内容")
    content_type: ContentType = Field(description="chunk类型", default=ContentType.TEXT)
    search_content: Optional[str] = Field(description="用来检索的内容", default=None)
    page_idx: int = Field(description="chunk在文档中的页码,从1开始")






class AbstractLoader:
    @abstractmethod
    def load(self, file_path: str) -> List[Chunk]:
        raise NotImplementedError



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/02/20 15:47:53
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''






from typing import List, Dict


import requests
from snippets import *
from xagents.model import LLM, EMBD

from loguru import logger

# model = SentenceTransformer(model_name_or_path='/Users/chenhao/Documents/zhipu/deploy/model/m3e-base')
MOCK_SN = "你好，这是一条mock信息"

def call_sn_chat(messages: List[Dict], temperature=0.01, top_p=0.7, max_new_token=1024, **kwargs):
    # data = dict(inputs=messages[0])
    # url = "http://10.0.251.202:8080/generate"
    # resp = requests.post(url=url, json=data)
    # resp.raise_for_status()
    # content = resp.json()["generated_text"]
    content = MOCK_SN
    return None, content

def call_sn_embedding(contents: List[str]):
    embeddings = []
    url = "http://localhost:8009/embd"
    resp = requests.post(url=url, data=dict(text=contents))
    resp.raise_for_status()
    return resp.json()["data"]

class SN_EMBD(EMBD):

    def __init__(self,  batch_size=16, norm=True):
        self.batch_size = batch_size
        self.norm = norm

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        logger.info(f"embedding {len(texts)} with {self.batch_size=}")

        embeddings = []
        from tqdm import tqdm
        
        for idx, batch in tqdm(enumerate(batchify(texts, self.batch_size))):
            try:
                embds = call_sn_embedding(contents=batch)
                embeddings.extend(embds)
                pct = idx*self.batch_size/len(texts)
                logger.info(f"{idx*self.batch_size}/{len(texts)} [{pct:2.2%}] embeddings done")
            except Exception as e:
                logger.error(f"embedding service error")
                logger.error(batch)

                raise e
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        embedding = call_sn_embedding(contents=[text])[0]
        return embedding

    @classmethod
    def get_dim(cls) -> int:
        return 768


class SN_LLM(LLM):
    @classmethod
    def list_versions(cls):
        return [
            "v1.0.0",
        ]

    def __init__(self, name="ZETA", version="v1.0.0"):
        super().__init__(name=name, version=version)

    def generate(self, prompt, history=[], stream=True,
                 temperature=0.01, **kwargs):
        messages = history + [dict(role="user", content=prompt)]
        response = call_sn_chat(messages=messages,
                                  temperature=temperature, **kwargs)
        resp_message = response
        return (e for e in resp_message) if stream else resp_message


if __name__ == "__main__":

    # zeta_llm = SN_LLM()
    # resp = zeta_llm.generate("你好", stream=False)
    # print(f"llm resp:{resp}")

    zeta_embd = SN_EMBD()
    contents = ["万得", "金融"]
    embds = zeta_embd.embed_documents(texts=contents)
    print(f"embd num {len(embds)}")
    print(f"embd dim {len(embds[0])}")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/01/04 14:04:02
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''

from xagents.tool.core import BaseTool, Parameter

    
def exec_python(expression:str)->dict:
    try:
        rs = eval(expression)
    except Exception as e:
        rs = "执行失败"
    return dict(result=rs)
    
    
calculator = BaseTool(name="计算器", description="根据提供的数学表达式，用python解释器来执行，得到计算结果,计算结果以json格式来返回,json包含一个字段，名字为result",
                      parameters=[Parameter(name="expression", type="string", description="数学表达式，可以通过python来执行的", required=True)],
                      callable=exec_python)
            
            
if __name__ == "__main__":
    rs = calculator.execute("(351345-54351)/54351")
    print(rs)
    
    
    
    
        



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/11 16:42:25
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''


from typing import List
from xagents.config import *
from xagents.loader.common import Chunk, ContentType, AbstractLoader
from loguru import logger


class PDFLoader(AbstractLoader):
    def __init__(self, max_page=None):
        self.max_page = max_page

    def load(self, file_path: str) -> List[Chunk]:
        import PyPDF2
        chunks = []
        with open(file_path, "rb") as f:
            pdf_reader = PyPDF2.PdfReader(f)
            logger.debug(f"got {len(pdf_reader.pages)} pages")
            pages = pdf_reader.pages
            if self.max_page:
                pages = pages[:self.max_page]
            for idx, page in enumerate(pages):
                chunks.append(Chunk(content=page.extract_text(), page_idx=idx+1, content_type=ContentType.TEXT))
        return chunks


if __name__ == "__main__":
    from xagents.config import XAGENT_HOME
    loader = PDFLoader()
    file_path = os.path.join(XAGENT_HOME, "data/raw/贵州茅台2022年报-4.pdf")
    chunks = loader.load(file_path)

    print(len(chunks))
    print(chunks[0])
